import importlib.util

if (
    importlib.util.find_spec("liquid") is None
    or importlib.util.find_spec("openai") is None
    or importlib.util.find_spec("tiktoken") is None
):
    raise ImportError(
        "not_again_ai.llm requires the 'llm' extra to be installed. "
        "You can install it using 'pip install not_again_ai[llm]'."
    )
else:
    import liquid  # noqa: F401
    import openai  # noqa: F401
    import tiktoken  # noqa: F401



import matplotlib
import matplotlib.dates
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import seaborn as sns

from not_again_ai.base.file_system import create_file_dir
from not_again_ai.viz.utils import reset_plot_libs


def ts_lineplot(
    ts_data: list[float] | (npt.NDArray[np.float64] | npt.NDArray[np.int64]),
    save_pathname: str,
    ts_x: (
        list[float]
        | (npt.NDArray[np.float64] | (npt.NDArray[np.datetime64] | (npt.NDArray[np.int64] | pd.Series)))
        | None
    ) = None,
    ts_names: list[str] | None = None,
    title: str | None = None,
    xlabel: str | None = "Time",
    ylabel: str | None = "Value",
    legend_title: str | None = None,
    xaxis_date_format: str | None = None,
    xaxis_major_locator: matplotlib.ticker.Locator | None = None,
    ylim: tuple[float, float] | None = None,
    yticks: npt.ArrayLike | None = None,
    font_size: float = 48,
    height: float = 13,
    aspect: float = 2.2,
    linewidth: float = 2,
    legend_loc: str | (tuple[float, float] | int) | None = None,
    palette: str | (list[str] | (list[float] | (dict[str, str] | matplotlib.colors.Colormap))) = "tab10",
) -> None:
    """Saves a time series plot where each row in `ts_data` is a time series.
    Optionally, a specific x axis (like dates) can be provided with `ts_x`.
    Names to appear in the legend for each time series can be provided with `ts_names`.

    Args:
        ts_data (list of lists or 2D numpy array): Each nested list or row is a time series to be plotted.
        save_pathname (str): Filepath to save plot to. Parent directories will be automatically created.
        ts_x (listlike, optional): The values that will be used for the x-axis. Defaults to None.
        ts_names (list[str], optional): The names of the time series shown on the legend. Defaults to None.
        title (str, optional): Title of the plot. Defaults to None.
        xlabel (str, optional): Set the label for the x-axis. Defaults to 'Time'.
        ylabel (str, optional): Set the label for the y-axis. Defaults to 'Value'.
        legend_title (str, optional): Sets the title of the legend. Defaults to None.
        xaxis_date_format (str, optional): A dateformat string. See [strftime cheatsheet](https://strftime.org/). Defaults to None.
        xaxis_major_locator (matplotlib.ticker.Locator, optional): Matplotlib tick locator,
            See [Tick locating](https://matplotlib.org/stable/api/ticker_api.html) or [Date tickers](https://matplotlib.org/stable/api/ticker_api.html) for the available options. Defaults to None.
        ylim (tuple[float, float], optional): Set the y-axis limits (lower, upper). Defaults to None.
        yticks (npt.ArrayLike, optional): Set the y-axis tick locations. Defaults to None.
        font_size (float, optional): Font size. Defaults to 48.
        height (float, optional): Height (in inches) of the plot. Defaults to 13.
        aspect (float, optional): Aspect ratio of the plot, so that `aspect` * `height` gives the width of each facet in inches. Defaults to 2.2.
        linewidth (float, optional): Size of each time series line. Defaults to 2.
        legend_loc (Union[str, tuple[float, float], int], optional): Matplotlib legend location.
            See [matplotlib documentation](https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html). Defaults to None.
        palette (str, list, dict, matplotlib.colors.Colormap], optional): Takes the same arguments as [seaborn's lineplot](https://seaborn.pydata.org/generated/seaborn.lineplot.html#seaborn-lineplot) palette argument.
    """

    sns.set_theme(
        style="white",
        rc={
            "font.size": font_size,
            "axes.titlesize": font_size,
            "axes.labelsize": font_size * 0.8,
            "xtick.labelsize": font_size * 0.65,
            "ytick.labelsize": font_size * 0.65,
            "legend.fontsize": font_size * 0.5,
            "legend.title_fontsize": font_size * 0.55,
        },
    )
    # Transpose the list of lists or numpy array
    ts_data = np.array(ts_data).T
    sns_data = pd.DataFrame(ts_data, columns=ts_names)
    if ts_x is None:
        ts_x = np.arange(len(ts_data))

    sns_data["Time"] = ts_x
    sns_data = sns_data.melt(id_vars="Time", var_name="Time Series", value_name="Value")
    ax = sns.lineplot(data=sns_data, x="Time", y="Value", hue="Time Series", palette=palette, linewidth=linewidth)

    ax.figure.set_size_inches(height * aspect, height)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.legend(title=legend_title)

    ax.set_ylim(ylim)

    if legend_loc is not None:
        ax.legend(loc=legend_loc)

    if (xaxis_date_format is not None) and (ts_x is not None):
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter(xaxis_date_format))

    if (xaxis_major_locator is not None) and (ts_x is not None):
        ax.xaxis.set_major_locator(xaxis_major_locator)

    if yticks is not None:
        ax.set(yticks=yticks)

    sns.despine(top=True, right=True)

    create_file_dir(save_pathname)
    plt.savefig(save_pathname, bbox_inches="tight")
    reset_plot_libs()



import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import seaborn as sns

from not_again_ai.base.file_system import create_file_dir
from not_again_ai.viz.utils import reset_plot_libs


def simple_barplot(
    x: list[str] | (list[float] | (npt.NDArray[np.int_] | npt.NDArray[np.float64])),
    y: list[str] | (list[float] | (npt.NDArray[np.int_] | npt.NDArray[np.float64])),
    save_pathname: str,
    order: str | None = None,
    orient_bars_vertically: bool = True,
    title: str | None = None,
    x_label: str | None = None,
    y_label: str | None = None,
    font_size: float = 48,
    height: float = 11,
    aspect: float = 2,
) -> None:
    """Saves a simple barplot to the specified pathname.

    Args:
        x (listlike): Input listlike data for x-axis.
            If orient_bars_vertically is True, this is the category names.
            If orient_bars_vertically is False, this is the bar heights (cannot be numeric).
        y (listlike): Input listlike data for y-axis
            If orient_bars_vertically is True, this is the bar heights (cannot be numeric).
            If orient_bars_vertically is False, this is the category names.
        save_pathname (str): Filepath to save plot to. Parent directories will be automatically created.
        order (str, optional): Order of the bars, either "asc" or "desc". Defaults to None.
        orient_bars_vertically (bool, optional): Whether to orient the bars vertically. Defaults to True.
        title (str, optional): Title of the plot. Defaults to None.
        x_label (str, optional): Label for the x-axis. Defaults to None.
        y_label (str, optional): Label for the y-axis. Defaults to None.
        font_size (float, optional): Font size. Defaults to 48.
        height (float, optional): Height (in inches) of the plot. Defaults to 11.
        aspect (float, optional): Aspect ratio of the plot. Defaults to 2.
    """

    sns.set_theme(
        style="white",
        rc={
            "font.size": font_size,
            "axes.titlesize": font_size,
            "axes.labelsize": font_size * 0.8,
            "xtick.labelsize": font_size * 0.65,
            "ytick.labelsize": font_size * 0.65,
            "legend.fontsize": font_size * 0.5,
            "legend.title_fontsize": font_size * 0.55,
        },
    )
    sns.set_color_codes("muted")

    if order:
        if order == "asc":
            # sort x and y ascending by y
            if orient_bars_vertically:
                y, x = (list(t) for t in zip(*sorted(zip(y, x, strict=True)), strict=True))
            else:
                x, y = (list(t) for t in zip(*sorted(zip(x, y, strict=True)), strict=True))
        elif order == "desc":
            # sort x and y descending by y
            if orient_bars_vertically:
                y, x = (list(t) for t in zip(*sorted(zip(y, x, strict=True), reverse=True), strict=True))
            else:
                x, y = (list(t) for t in zip(*sorted(zip(x, y, strict=True), reverse=True), strict=True))

    ax = sns.barplot(x=x, y=y, color="b", orient="v" if orient_bars_vertically else "h")
    ax.figure.set_size_inches(height * aspect, height)

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    sns.despine()

    create_file_dir(save_pathname)
    plt.savefig(save_pathname, bbox_inches="tight")
    reset_plot_libs()



from collections.abc import Callable
import json
import os
import re
import time
from typing import Any, Literal, cast

from loguru import logger
from ollama import ChatResponse, Client, ResponseError

from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Function,
    ToolCall,
)

OLLAMA_PARAMETER_MAP = {
    "frequency_penalty": "repeat_penalty",
    "max_completion_tokens": "num_predict",
    "context_window": "num_ctx",
    "n": None,
    "tool_choice": None,
    "reasoning_effort": None,
    "parallel_tool_calls": None,
    "logit_bias": None,
    "top_logprobs": None,
    "presence_penalty": None,
}


def validate(request: ChatCompletionRequest) -> None:
    if request.json_mode and request.structured_outputs is not None:
        raise ValueError("json_schema and json_mode cannot be used together.")

    # Check if any of the parameters set to OLLAMA_PARAMETER_MAP are not None
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is None and getattr(request, key) is not None:
            logger.warning(f"Parameter {key} is not supported by Ollama and will be ignored.")

    # If "stop" is not None, check if it is just a string
    if isinstance(request.stop, list):
        logger.warning("Parameter 'stop' needs to be a string and not a list. It will be ignored.")
        request.stop = None


def ollama_chat_completion(
    request: ChatCompletionRequest,
    client: Callable[..., Any],
) -> ChatCompletionResponse:
    validate(request)

    kwargs = request.model_dump(mode="json", exclude_none=True)

    # For each key in OLLAMA_PARAMETER_MAP
    # If it is not None, set the key in kwargs to the value of the corresponding value in OLLAMA_PARAMETER_MAP
    # If it is None, remove that key from kwargs
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is not None and key in kwargs:
            kwargs[value] = kwargs.pop(key)
        elif value is None and key in kwargs:
            del kwargs[key]

    # If json_mode is True, set the format to json
    json_mode = kwargs.get("json_mode", None)
    if json_mode:
        kwargs["format"] = "json"
        kwargs.pop("json_mode")
    elif json_mode is not None and not json_mode:
        kwargs.pop("json_mode")

    # If structured_outputs is not None, set the format to structured_outputs
    if kwargs.get("structured_outputs", None):
        # Check if the schema is in the OpenAI and pull out the schema
        if "schema" in kwargs["structured_outputs"]:
            kwargs["format"] = kwargs["structured_outputs"]["schema"]
            kwargs.pop("structured_outputs")
        else:
            kwargs["format"] = kwargs.pop("structured_outputs")

    option_fields = [
        "mirostat",
        "mirostat_eta",
        "mirostat_tau",
        "num_ctx",
        "repeat_last_n",
        "repeat_penalty",
        "temperature",
        "seed",
        "stop",
        "tfs_z",
        "num_predict",
        "top_k",
        "top_p",
        "min_p",
    ]
    # For each field in option_fields, if it is in kwargs, make it under an options dictionary
    options = {}
    for field in option_fields:
        if field in kwargs:
            options[field] = kwargs.pop(field)
    kwargs["options"] = options

    for message in kwargs["messages"]:
        role = message.get("role", None)
        # For each ToolMessage, remove the name field
        if role is not None and role == "tool":
            message.pop("name")

        # For each AssistantMessage with tool calls, remove the id field
        if role is not None and role == "assistant" and message.get("tool_calls", None):
            for tool_call in message["tool_calls"]:
                tool_call.pop("id")

        # Content and images need to be separated
        images = []
        content = ""
        if isinstance(message["content"], list):
            for item in message["content"]:
                if item["type"] == "image_url":
                    image_url = item["image_url"]["url"]
                    # Remove the data URL prefix if present
                    if image_url.startswith("data:"):
                        image_url = image_url.split("base64,", 1)[1]
                    images.append(image_url)
                else:
                    content += item["text"]
        else:
            content = message["content"]

        message["content"] = content
        if len(images) > 1:
            images = images[:1]
            logger.warning("Ollama model only supports a single image per message. Using only the first images.")
        message["images"] = images

    try:
        start_time = time.time()
        response: ChatResponse = client(**kwargs)
        end_time = time.time()
        response_duration = round(end_time - start_time, 4)
    except ResponseError as e:
        # If the error says "model 'model' not found" use regex then raise a more specific error
        expected_pattern = f"model '{request.model}' not found"
        if re.search(expected_pattern, e.error):
            raise ResponseError(f"Model '{request.model}' not found.") from e
        else:
            raise ResponseError(e.error) from e

    errors = ""

    # Handle tool calls
    tool_calls: list[ToolCall] | None = None
    if response.message.tool_calls:
        parsed_tool_calls: list[ToolCall] = []
        for tool_call in response.message.tool_calls:
            tool_name = tool_call.function.name
            if request.tools and tool_name not in [tool["function"]["name"] for tool in request.tools]:
                errors += f"Tool call {tool_call} has an invalid tool name: {tool_name}\n"
            tool_args = tool_call.function.arguments
            parsed_tool_calls.append(
                ToolCall(
                    id="",
                    function=Function(
                        name=tool_name,
                        arguments=tool_args,
                    ),
                )
            )
        tool_calls = parsed_tool_calls

    json_message = None
    if (request.json_mode or (request.structured_outputs is not None)) and response.message.content:
        try:
            json_message = json.loads(response.message.content)
        except json.JSONDecodeError:
            errors += "Message failed to parse into JSON\n"

    finish_reason = cast(
        Literal["stop", "length", "tool_calls", "content_filter"],
        "stop" if response.done_reason is None else response.done_reason or "stop",
    )

    choice = ChatCompletionChoice(
        message=AssistantMessage(
            content=response.message.content or "",
            tool_calls=tool_calls,
        ),
        finish_reason=finish_reason,
        json_message=json_message,
    )

    return ChatCompletionResponse(
        choices=[choice],
        errors=errors.strip(),
        completion_tokens=response.get("eval_count", -1),
        prompt_tokens=response.get("prompt_eval_count", -1),
        response_duration=response_duration,
    )


def ollama_client(host: str | None = None, timeout: float | None = None) -> Callable[..., Any]:
    """Create an Ollama client instance based on the specified host or will read from the OLLAMA_HOST environment variable.

    Args:
        host (str, optional): The host URL of the Ollama server.
        timeout (float, optional): The timeout for requests

    Returns:
        Client: An instance of the Ollama client.

    Examples:
        >>> client = client(host="http://localhost:11434")
    """
    if host is None:
        host = os.getenv("OLLAMA_HOST")
        if host is None:
            logger.warning("OLLAMA_HOST environment variable not set, using default host: http://localhost:11434")
            host = "http://localhost:11434"

    def client_callable(**kwargs: Any) -> Any:
        client = Client(host=host, timeout=timeout)
        return client.chat(**kwargs)

    return client_callable



import matplotlib.pyplot as plt
import seaborn as sns


def reset_plot_libs() -> None:
    """Resets the plot libraries so that subsequent method calls are not impacted."""
    plt.clf()
    plt.cla()
    plt.close()
    sns.reset_orig()



import importlib.util

if (
    importlib.util.find_spec("numpy") is None
    or importlib.util.find_spec("scipy") is None
    or importlib.util.find_spec("sklearn") is None
):
    raise ImportError(
        "not_again_ai.statistics requires the 'statistics' extra to be installed. "
        "You can install it using 'pip install not_again_ai[statistics]'."
    )
else:
    import numpy  # noqa: F401
    import scipy  # noqa: F401
    import sklearn  # noqa: F401



from not_again_ai.llm.chat_completion.interface import chat_completion
from not_again_ai.llm.chat_completion.types import ChatCompletionRequest

__all__ = ["ChatCompletionRequest", "chat_completion"]



from pathlib import Path


def create_file_dir(filepath: str | Path) -> None:
    """Creates the parent directories for the specified filepath.
    Does not throw any errors if the directories already exist.

    Args:
        filepath (str | Path): path to a file
    """
    root_path = Path(filepath).parent
    root_path.mkdir(parents=True, exist_ok=True)


def readable_size(size: float) -> str:
    """Convert a file size given in bytes to a human-readable format.

    Args:
        size (int): file size in bytes

    Returns:
        str: human-readable file size
    """
    # Define the suffixes for each size unit
    suffixes = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]

    # Start with bytes
    count = 0
    while size >= 1024 and count < len(suffixes) - 1:
        count += 1
        size /= 1024

    # Format the size to two decimal places and append the appropriate suffix
    return f"{size:.2f} {suffixes[count]}"



from collections.abc import Callable
from typing import Any

from not_again_ai.llm.embedding.providers.ollama_api import ollama_create_embeddings
from not_again_ai.llm.embedding.providers.openai_api import openai_create_embeddings
from not_again_ai.llm.embedding.types import EmbeddingRequest, EmbeddingResponse


def create_embeddings(request: EmbeddingRequest, provider: str, client: Callable[..., Any]) -> EmbeddingResponse:
    """Get a embedding response from the given provider. Currently supported providers:
    - `openai` - OpenAI
    - `azure_openai` - Azure OpenAI
    - `ollama` - Ollama

    Args:
        request: Request parameter object
        provider: The supported provider name
        client: Client information, see the provider's implementation for what can be provided

    Returns:
        EmbeddingResponse: The embedding response.
    """
    if provider == "openai" or provider == "azure_openai":
        return openai_create_embeddings(request, client)
    elif provider == "ollama":
        return ollama_create_embeddings(request, client)
    else:
        raise ValueError(f"Provider {provider} not supported")
